/**
 * @file evm_subroutines.c
 * @author Joao Paulo Martins (joaopaulo.martins@ess.eu)
 * @author Gabriel Fedel (gabriel.fedel@ess.eu)
 * @brief 
 * @version 0.1
 * @date 2020-07-17
 * 
 * @copyright Copyright (c) ESS 2020
 * 
 * The code below is based on supercycleEngine project:
 * https://github.com/icshwi/supercycleengine
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <aSubRecord.h>
#include <registryFunction.h>
#include <epicsExport.h>
#include <epicsTypes.h>

#include <string.h>

//------------------------------------------------------------------------
//  Event Codes
//------------------------------------------------------------------------
#define EVT14HZ     14
#define RF_ST       60
#define RF_END      61
#define BPULSE_CM   15
#define BPULSE_ST   12
#define BPULSE_END  13
#define SEQ_END     127

#define MAX_SEQ_EVENTS 6
#define RF_FREQUENCY 88052500

//------------------------------------------------------------------------
//  Routine that writes the EVM sequencer events and timestamps arrays 
//------------------------------------------------------------------------
static long seqConfigure(aSubRecord *prec)
{
    int   *event_codes;
    double *timestamps;

    // Get RF pulse length from INPA
    float rflen = *(float *)prec->a;
    
    // Get Proton Beam pulse length from INPB
    float pblen = *(float *)prec->b;

    // Get delays
    float rfdelay = 10000.0; // make constant to avoid rep. rate interlocks
    float pbdelay = *(float *)prec->c;

    // Get frequency (period)
    float cyclePeriod = *(float *)prec->d;
    cyclePeriod = (1.0 / cyclePeriod) * 1000000.0; // period in microseconds
    float maxRFDelay = cyclePeriod - 10000.0;

    // Get Pulse Coming Delay
    float pulsecm = *(float *)prec->e;

    // Validate RF delay and RF length
    rfdelay = (rfdelay < 1.0) ? 1.0 : rfdelay;
    rfdelay = (rfdelay >= maxRFDelay ? maxRFDelay : rfdelay); 
    rflen = ((rfdelay + rflen) > (maxRFDelay + 404.0)) ? (maxRFDelay + 404.0 - rfdelay) : rflen;
    
    // Validate Pulse Coming Delay
    pulsecm = (pulsecm < 100.0) ? 100.0 : pulsecm;
    pulsecm = ((rflen - pulsecm) < 302.0) ? (rflen - 302.0) : pulsecm;

    // Validate PB Pulse
    pbdelay = ((pbdelay - pulsecm) < 300.0) ? pulsecm + 300.0 : pbdelay;
    pbdelay = (pbdelay > (rflen - 2.0)) ? (rflen - 2.0) : pbdelay;
    pblen = (pblen >= (rflen - pbdelay)) ? (rflen - pbdelay - 1.0) : pblen;

    event_codes = (int*) malloc(MAX_SEQ_EVENTS * sizeof(int));
    timestamps = (double*) malloc(MAX_SEQ_EVENTS * sizeof(double));
    
    // Create the array of events
    event_codes[0] = RF_ST;
    event_codes[1] = BPULSE_CM;
    event_codes[2] = BPULSE_ST;
    event_codes[3] = BPULSE_END;
    event_codes[4] = RF_END;
    event_codes[5] = SEQ_END;
    
    // Create the array of time stamps
    timestamps[0] = rfdelay;                    // RF_ST
    timestamps[1] = timestamps[0] + pulsecm;    // BPULSE_CM
    timestamps[2] = timestamps[0] + pbdelay;    // BPULSE_ST
    timestamps[3] = timestamps[2] + pblen;      // BPULSE_END
    timestamps[4] = timestamps[0] + rflen;      // RF_END
    timestamps[5] = timestamps[4] + 1.0;        // sequencer re-arm

    // Write the output arrays
    memcpy(prec->vala, event_codes, MAX_SEQ_EVENTS * sizeof(int));
    prec->neva = MAX_SEQ_EVENTS;
    memcpy(prec->valb, timestamps, MAX_SEQ_EVENTS * sizeof(double));
    prec->nevb = MAX_SEQ_EVENTS;

    // Write the readback legth PVs
    memcpy(prec->valc, &rflen, sizeof(float));
    prec->nevc = 1;
    memcpy(prec->vald, &pblen, sizeof(float));
    prec->nevd = 1;
    memcpy(prec->vale, &rfdelay, sizeof(float));
    prec->neve = 1;
    memcpy(prec->valf, &pbdelay, sizeof(float));
    prec->nevf = 1;
    memcpy(prec->valg, &pulsecm, sizeof(float));
    prec->nevg = 1;
    
    free(event_codes);
    free(timestamps);

    return 0;
}
epicsRegisterFunction(seqConfigure);


